/*=============================================================================
	Collision Cacus header
=============================================================================*/

#ifndef _INC_XC_COL
#define _INC_XC_COL

#ifndef DISABLE_COLLISION

#define MAX_OCTREE_DEPTH 6
#define OCTREE_GRID_ARRAY 7
#define TWO_EXP_OGA 64 //2^OCTREE_GRID_ARRAY

#define MIN_ACTORS_FOR_SUBNODE 3

class FPrecomputedRay
{
public:
	FVector Org;
	FVector End;
	FVector Dir;
	FVector Inv;
	FBox ExtendedBox; //For grid interaction and quick rejects
	FVector XAxis; //For CylCyl checks
	FVector YAxis;
	const FVector* Extent;
	FLOAT Dist;
	UBOOL(FPrecomputedRay::*TraceFunc)( AActor*, FVector*, FVector*);

	FPrecomputedRay()	{};
	FPrecomputedRay( const FVector& A, const FVector& B, const FVector& TExtent);


	inline UBOOL FitsBox( const FVector& Sample, const FBox& Container)
	{
		return (Sample.X >= Container.Min.X) && (Sample.Y >= Container.Min.Y) && (Sample.Z >= Container.Min.Z)
			&& (Sample.X <= Container.Max.X) && (Sample.Y <= Container.Max.Y) && (Sample.Z <= Container.Max.Z);
	}

	void TwoPointBox( FBox& Box, const FVector& A, const FVector& B);
	INT IntersectsBox( const FBox& Box);

	UBOOL HitsGCylActor( AActor* Actor, FVector* HitLocation, FVector* HitNormal);
	UBOOL HitsVCylActor( AActor* Actor, FVector* HitLocation, FVector* HitNormal);
	UBOOL HitsCylActor( AActor* Actor, FVector* HitLocation, FVector* HitNormal);
};

struct XC_ENGINE_API FCacusActorInfo
{
	AActor* Actor;
	INT SubOctant; //-1 if not
	FBox Box;
	FVector Location;
	FCacusActorInfo( AActor* NewActor, INT NewOctant=-1)
	:	Actor(NewActor)
	,	SubOctant(NewOctant)
	,	Box( NewActor->GetPrimitive()->GetCollisionBoundingBox(NewActor) )
	,	Location( NewActor->Brush ? ((Box.Min+Box.Max)*0.5) : NewActor->Location)
	{
		if ( NewActor->Brush )
			Box.ExpandBy( 1.f);
	}
};

class XC_ENGINE_API FCollisionOctree
{
public:
	FCollisionCacus* Master;
	FCollisionOctree* Parent;
	FCollisionOctree* Children[8]; //X,Y,Z based bitwise.
	FCollisionOctree* ReducedChildren[8]; //For fast operations
	FBox BoundingBox;
	FVector CenterPoint;
	BYTE CurDepth;
	BYTE ChildIndex;
	BYTE ChildCount;
	BYTE CleanupTimer;
	BYTE XPos, YPos, ZPos;
	INT MIndex; //Index on master array
	TArray<FCacusActorInfo> Actors;
	TArray<FCacusActorInfo> BrushMovers; //Brush && bIsMover
	//FCollisionCacus uses this
	FCollisionOctree()
	:	CurDepth(0)
	,	ChildIndex(0)
	,	ChildCount(0)
	,	CleanupTimer(20)
	,	Actors(0)
	,	BrushMovers(0)
	,	CenterPoint(0,0,0)
	,	XPos(0)	,	YPos(0)	,	ZPos(0)
	,	BoundingBox( FVector(-32768,-32768,-32768), FVector(32768,32768,32768))
	{
		appMemzero( &Parent, 4*17);
	}
	FCollisionOctree( FCollisionOctree* InParent, FBox InBounds, INT InDepth, BYTE InChild=0)
	:	Parent(InParent)
	,	BoundingBox(InBounds)
	,	CenterPoint( (InBounds.Min+InBounds.Max) * 0.5)
	,	CurDepth(InDepth)
	,	ChildIndex( InChild)
	,	CleanupTimer(20)
	,	ChildCount(0)
	,	XPos(0)	,	YPos(0)	,	ZPos(0)
	,	Actors(0)
	,	BrushMovers(0)
	{
		appMemzero( &Children[0], 4*16 ); //Both child arrays
		if ( InParent )
		{
#ifdef CACUS_DEBUG
			check( InParent->ChildCount < 8 );
			check( !InParent->ReducedChildren[InParent->ChildCount] );
#endif
			InParent->ReducedChildren[InParent->ChildCount] = this;
			InParent->ChildCount++;
		}
#ifdef CACUS_DEBUG
		check( InBounds.Min.X < InBounds.Max.X );
		check( InBounds.Min.Y < InBounds.Max.Y );
		check( InBounds.Min.Z < InBounds.Max.Z );
//		debugf( NAME_Log, TEXT("[XC_GE] Creating new Octree node, Depth %i"), CurDepth);
#endif
	}
	~FCollisionOctree();
	virtual void Exit();

	void Cleanup();
	FBox GetSubOctantBox( const FVector& Point, INT& Idx);		//Fast, does not do sanity checks
	INT GetSubOctantIdx( const FVector& Point);				//Same
	FCacusActorInfo* FindActor( FName ByName);
	FIteratorActorList* RadiusQuery( FMemStack& Mem, const FVector& Location, const FLOAT& Radius, const DWORD& ExtraNodeFlags, FIteratorActorList* PrevLink );
	FCheckResult* PointQuery( FMemStack& Mem, const FVector& Location, const FVector& Extent, const DWORD& ExtraNodeFlags, FCheckResult* PrevLink );
	FCheckResult* FastPointQuery( FMemStack& Mem, const FVector& Location, const FVector& Extent, const DWORD& ExtraNodeFlags, FCheckResult* PrevLink);
	FCheckResult* MEncroachQuery( FMemStack& Mem, AActor* Actor, const FBox& ActorBox, FCheckResult* PrevLink=NULL);
	FCheckResult* PEncroachQuery( FMemStack& Mem, AActor* Actor, const FVector& Location, const FBox& ActorBox, FCheckResult* PrevLink=NULL);
	FCheckResult* AEncroachQuery( FMemStack& Mem, AActor* Actor, const FVector& Location, const FBox& ActorBox, FCheckResult* PrevLink=NULL);
	FCheckResult* LineQuery
	(
	 FMemStack& Mem
	, const FVector& End
	, const FVector& Start
	, const FVector& Extent
	, const BYTE& ExtraNodeFlags
	, FPrecomputedRay& Ray
	,  FCheckResult* PrevLink
	);
	FCheckResult* FastLineQuery
	(
	 FMemStack& Mem
	, const FVector& End
	, const FVector& Start
	, const FVector& Extent
	, const BYTE& ExtraNodeFlags
	, FPrecomputedRay& Ray
	, FCheckResult* PrevLink
	);
	inline UBOOL FitsBox( const FBox& Sample, const FBox& Container)
	{
		return (Sample.Min.X >= Container.Min.X) && (Sample.Min.Y >= Container.Min.Y) && (Sample.Min.Z >= Container.Min.Z)
			&& (Sample.Max.X <= Container.Max.X) && (Sample.Max.Y <= Container.Max.Y) && (Sample.Max.Z <= Container.Max.Z);
	}
	inline UBOOL FitsBox( const FVector& Sample, const FBox& Container)
	{
		return (Sample.X >= Container.Min.X) && (Sample.Y >= Container.Min.Y) && (Sample.Z >= Container.Min.Z)
			&& (Sample.X < Container.Max.X) && (Sample.Y < Container.Max.Y) && (Sample.Z < Container.Max.Z);
	}
	inline UBOOL FitsBox( AActor* Actor, const FBox& Container)
	{
		return (Actor->Location.X - Actor->CollisionRadius >= Container.Min.X) && (Actor->Location.Y - Actor->CollisionRadius >= Container.Min.Y) && (Actor->Location.Z - Actor->CollisionHeight >= Container.Min.Z)
			&& (Actor->Location.X + Actor->CollisionRadius < Container.Max.X) && (Actor->Location.Y + Actor->CollisionRadius < Container.Max.Y) && (Actor->Location.Z + Actor->CollisionHeight < Container.Max.Z);
	}
	inline FBox ExpandByExtent( const FBox& Box, const FVector& Extent)
	{	return FBox( Box.Min-Extent, Box.Max+Extent);	}
	//Quick reject, actor should not go into a sub-box
	inline UBOOL HitsCenterAxes( const FBox& Box)
	{
		return ( ((Box.Min.X < CenterPoint.X) && (Box.Max.X > CenterPoint.X))
		||	((Box.Min.Y < CenterPoint.Y) && (Box.Max.Y > CenterPoint.Y))
		||	((Box.Min.Z < CenterPoint.Z) && (Box.Max.Z > CenterPoint.Z)) );
	}
};

//====================================
// Three dimension grid
//==========

//V451 is cancer, i need to work around it by statically allocating a struct
struct FGridData
{
	INT Position[OCTREE_GRID_ARRAY];

	FCollisionOctree* Grid0;
#if OCTREE_GRID_ARRAY > 0
	FCollisionOctree* Grid1[2][2][2];
#endif
#if OCTREE_GRID_ARRAY > 1
	FCollisionOctree* Grid2[4][4][4];
#endif
#if OCTREE_GRID_ARRAY > 2
	FCollisionOctree* Grid3[8][8][8];
#endif
#if OCTREE_GRID_ARRAY > 3
	FCollisionOctree* Grid4[16][16][16];
#endif
#if OCTREE_GRID_ARRAY > 4
	FCollisionOctree* Grid5[32][32][32];
#endif
#if OCTREE_GRID_ARRAY > 5
	FCollisionOctree* Grid6[64][64][64];
#endif
#if OCTREE_GRID_ARRAY > 6
	FCollisionOctree* Grid7[128][128][128];
#endif
#if OCTREE_GRID_ARRAY > 7
	FCollisionOctree* Grid8[256][256][256];
#endif

	FGridData()
	{
		Position[0] = 0;
		for ( INT i=1 ; i<OCTREE_GRID_ARRAY ; i++ )
		{
			INT PrevSize = (1 << (i-1));
			PrevSize = PrevSize * PrevSize * PrevSize;
			Position[i] = Position[i-1] + PrevSize;
			INT NewSize = (1 << i);
			appMemzero( &Position[i], NewSize * NewSize * NewSize * 4);
		}
	}
};


class XC_ENGINE_API FNodeGrid3
{
public:
	INT GridSize;
	FCollisionOctree** Grid;
	//	INT Sub[3]; //FPU is causing lots of precision issues
	FVector Substract; //Substract this when calculating grid position, should be identical in all grids?
	FVector BoxSize; //Dividing a vector by this (post substraction) will give coordinates!
	FVector Divider; //Precalculated version

	FNodeGrid3()
	:	GridSize(0)
	,	Grid(NULL)
	{};

	void SetBaseGrid( FCollisionOctree* Base);
	void SetGrid( INT NewGridSize, const FBox& Bounds);
	void Set( FCollisionOctree* In);
	void Remove( FCollisionOctree* In);
	FCollisionOctree* Get( FVector Pos);
};

class XC_ENGINE_API FCollisionCacus : public FCollisionHashBase , public FCollisionOctree
{
public:
	FCollisionHashBase* Hooked;
	INT MinAvailable;
	TArray<FCollisionOctree*> ColList;
	AActor* JustRemoved;
	FGridData* GridData;
	FNodeGrid3 DepthGrid[OCTREE_GRID_ARRAY]; //Zero isn't used, we're doing this for quick CurDepth access

	FCollisionCacus()
	:	ColList(0)
	,	MinAvailable(0)
	,	GridData(NULL)
	{
		ColList.AddItem( this);
		MIndex = 0;
		Master = this;
		appMemzero( &DepthGrid, sizeof(class FNodeGrid3) * OCTREE_GRID_ARRAY );
	};
	~FCollisionCacus();
	void Exit();

	void SetupGrids();
	void ChildSet( FCollisionOctree* Child);
	void Tick();
	void AddActor( AActor *Actor );
	void RemoveActor( AActor *Actor );
	FCheckResult* ActorLineCheck( FMemStack& Mem, FVector End, FVector Start, FVector Extent, BYTE ExtraNodeFlags );
	FCheckResult* ActorPointCheck( FMemStack& Mem, FVector Location, FVector Extent, DWORD ExtraNodeFlags );
	FCheckResult* ActorRadiusCheck( FMemStack& Mem, FVector Location, FLOAT Radius, DWORD ExtraNodeFlags );
	FCheckResult* ActorEncroachmentCheck( FMemStack& Mem, AActor* Actor, FVector Location, FRotator Rotation, DWORD ExtraNodeFlags );
	void CheckActorNotReferenced( AActor* Actor );
};
#endif

#endif

/*-----------------------------------------------------------------------------
	The End.
-----------------------------------------------------------------------------*/
